from pathlib import Path
from typing import Any

from exponent.core.remote_execution import files
from exponent.core.remote_execution.types import (
    CommandRequest,
    CommandResponse,
)
from exponent.core.types.command_data import (
    FileReadCommandData,
    PrototypeCommandData,
)


# Intentionally split the separator into two parts
# to avoid matching it in the content
CONTEXT_BATCH_SEPARATOR = "\n<batch_sep" + "arator>\n"
CONTEXT_FILE_SEPARATOR = "\n<file_sep" + "arator>\n"


async def execute_command(
    request: CommandRequest,
    working_directory: str,
) -> CommandResponse:
    try:
        if isinstance(request.data, FileReadCommandData):
            correlation_id = request.correlation_id
            file_path = request.data.file_path
            path = Path(working_directory, file_path)
            content, _ = await files.get_file_content(path)

            return CommandResponse(
                subcommand=request.data.type.value,
                content=content,
                correlation_id=correlation_id,
            )
        elif isinstance(request.data, PrototypeCommandData):
            correlation_id = request.correlation_id
            command_name = request.data.command_name
            content_json = request.data.content_json
            content_raw = request.data.content_raw
            content_rendered = request.data.content_rendered

            content = await execute_prototype_command(
                command_name=command_name,
                content_json=content_json,
                content_raw=content_raw,
                content_rendered=content_rendered,
                working_directory=working_directory,
            )

            return CommandResponse(
                subcommand=command_name,
                content=content,
                correlation_id=correlation_id,
            )
        else:
            raise ValueError(f"Unknown command request: {request}")
    except Exception as e:  # noqa: BLE001 - TODO (Josh): Specialize errors for execution
        return CommandResponse(
            content="An error occurred during command execution: " + str(e),
            correlation_id=request.correlation_id,
        )


async def execute_prototype_command(
    command_name: str,
    content_json: dict[str, Any],
    content_raw: str,
    content_rendered: str,
    working_directory: str,
) -> str:
    if command_name == "file_open":
        return f'Successfully opened file "{content_json["file_path"]}"'
    elif command_name == "search_files":
        results = await files.search_files(
            path_str=content_json["path"],
            file_pattern=content_json["file_pattern"],
            regex=content_json["regex"],
            working_directory=working_directory,
        )
        return "\n".join(results)
    elif command_name == "codebase_context":
        batches = await files.get_all_file_contents(
            working_directory=working_directory,
        )

        return CONTEXT_BATCH_SEPARATOR.join(
            CONTEXT_FILE_SEPARATOR.join(batch) for batch in batches
        )
    elif command_name == "ls":
        results = await files.file_walk(
            directory=content_json["path"],
            ignore_extra=files.DEFAULT_IGNORES,
            max_files=1000,
        )
        return "\n".join(results)
    elif command_name == "glob":
        results = await files.glob(
            path=content_json["path"],
            glob_pattern=content_json["glob"],
        )
        return "\n".join(results)
    raise ValueError(f"Unhandled prototype command: {command_name}")



import datetime
import json
from datetime import timezone
from enum import Enum
from functools import cached_property
from pathlib import Path, PurePath
from typing import Annotated, Any, ClassVar, Generic, Literal, Optional, TypeVar, Union
from anyio import Path as AsyncPath
from exponent.core.remote_execution.error_info import SerializableErrorInfo
from exponent.core.types.event_types import FileWriteStrategyName
from pydantic import BaseModel, Field

from exponent.core.types.command_data import (
    CommandDataType,
)


class UseToolsMode(str, Enum):
    read_only = "read_only"
    read_write = "read_write"
    disabled = "disabled"


class UseToolsConfig(BaseModel):
    mode: UseToolsMode = UseToolsMode.read_only

    @property
    def read_only(self) -> bool:
        return self.mode == UseToolsMode.read_only

    @property
    def read_write(self) -> bool:
        return self.mode == UseToolsMode.read_write

    def disable(self) -> None:
        self.mode = UseToolsMode.disabled


class CreateChatResponse(BaseModel):
    chat_uuid: str


class StartChatRequest(BaseModel):
    chat_uuid: str
    prompt: str
    use_tools_config: UseToolsConfig


class StartChatResponse(BaseModel):
    chat_uuid: str


class RunWorkflowRequest(BaseModel):
    chat_uuid: str
    workflow_id: str


class ExecutionEndResponse(BaseModel):
    execution_ended: bool


class SignalType(str, Enum):
    disconnect = "disconnect"

    def __str__(self) -> str:
        return self.value


class GitInfo(BaseModel):
    branch: str
    remote: Optional[str]


class PythonEnvInfo(BaseModel):
    interpreter_path: Optional[str]
    interpreter_version: Optional[str]
    name: Optional[str] = "exponent"
    provider: Optional[Literal["venv", "pyenv", "pipenv", "conda"]] = "pyenv"


class SystemInfo(BaseModel):
    name: str
    cwd: str
    os: str
    shell: str
    git: Optional[GitInfo]
    python_env: Optional[PythonEnvInfo]


class HeartbeatInfo(BaseModel):
    exponent_version: Optional[str] = None
    editable_installation: bool = False
    system_info: Optional[SystemInfo]
    timestamp: datetime.datetime = Field(
        default_factory=lambda: datetime.datetime.now(timezone.utc)  # noqa: UP017
    )
    timestamp_received: Optional[datetime.datetime] = None


class RemoteFile(BaseModel):
    file_path: str
    working_directory: str = "."

    @cached_property
    def pure_path(self) -> PurePath:
        return PurePath(self.working_directory, self.file_path)

    @cached_property
    def path(self) -> Path:
        return Path(self.working_directory, self.file_path)

    @cached_property
    def name(self) -> str:
        return self.pure_path.name

    @cached_property
    def absolute_path(self) -> str:
        return self.path.absolute().as_posix()

    async def resolve(self, client_working_directory: str) -> str:
        working_directory = AsyncPath(self.working_directory, self.file_path)

        if not working_directory.is_absolute():
            working_directory = AsyncPath(client_working_directory, working_directory)

        return str(await working_directory.resolve())

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, RemoteFile):
            return False

        return self.path.name == other.path.name

    def __lt__(self, other: "RemoteFile") -> bool:
        # Prefer shorter paths
        if (cmp := self._cmp_path_len(other)) is not None:
            return cmp

        # Prefer paths sorted by parent directory
        if (cmp := self._cmp_path_str(other)) is not None:
            return cmp

        # Prefer paths with alphabetical first character
        return self._cmp_first_char(other)

    def __hash__(self) -> int:
        return hash(self.absolute_path)

    def _cmp_first_char(self, other: "RemoteFile") -> bool:
        return self._cmp_str(self.path.name, other.path.name)

    def _cmp_path_len(self, other: "RemoteFile") -> Optional[bool]:
        self_parts = self.path.absolute().parent.parts
        other_parts = other.path.absolute().parent.parts

        if len(self_parts) == len(other_parts):
            return None

        return len(self_parts) < len(other_parts)

    def _cmp_path_str(self, other: "RemoteFile") -> Optional[bool]:
        self_parts = self.path.absolute().parent.parts
        other_parts = other.path.absolute().parent.parts

        if self_parts == other_parts:
            return None

        for a, b in zip(self_parts, other_parts):
            if a != b:
                return self._cmp_str(a, b)

        return False

    @staticmethod
    def _cmp_str(s1: str, s2: str) -> bool:
        if s1[:1].isalpha() == s2[:1].isalpha():
            return s1 < s2

        return s1[:1].isalpha()


class URLAttachment(BaseModel):
    attachment_type: Literal["url"] = "url"
    url: str
    content: str


class FileAttachment(BaseModel):
    attachment_type: Literal["file"] = "file"
    file: RemoteFile
    content: str


MessageAttachment = Annotated[
    Union[FileAttachment, URLAttachment],
    Field(discriminator="attachment_type"),
]


Direction = Literal[
    "request",
    "response",
]

Namespace = Literal[
    "code_execution",
    "streaming_code_execution",
    "streaming_code_execution_chunk",
    "file_write",
    "command",
    "list_files",
    "get_file_attachment",
    "get_file_attachments",
    "get_matching_files",
    "system_context",
    "get_all_tracked_files",
    "halt",
    "switch_cli_chat",
    "error",
    "create_checkpoint",
    "rollback_to_checkpoint",
]

ErrorType = Literal["unknown_request_type", "request_error"]

SupportedLanguage = Literal[
    "python",
    "shell",
]

SUPPORTED_LANGUAGES: list[SupportedLanguage] = ["python", "shell"]


class RemoteExecutionMessageData(BaseModel):
    namespace: Namespace
    direction: Direction
    message_data: str

    def message_type(self) -> str:
        return f"{self.namespace}.{self.direction}"


class RemoteExecutionMessage(BaseModel):
    direction: ClassVar[Direction]
    namespace: ClassVar[Namespace]
    correlation_id: str

    @classmethod
    def message_type(cls) -> str:
        return f"{cls.namespace}.{cls.direction}"

    @property
    def result_key(self) -> str:
        return f"{self.namespace}:{self.correlation_id}"


### Response Types


class RemoteExecutionResponseData(RemoteExecutionMessageData):
    pass


class RemoteExecutionResponse(RemoteExecutionMessage):
    direction: ClassVar[Direction] = "response"


ResponseT = TypeVar("ResponseT", bound=RemoteExecutionResponse)


class StreamingCodeExecutionResponseChunk(RemoteExecutionResponse):
    namespace: ClassVar[Namespace] = "streaming_code_execution_chunk"

    content: str

    def add(
        self, new_chunk: "StreamingCodeExecutionResponseChunk"
    ) -> "StreamingCodeExecutionResponseChunk":
        """Aggregates content of this and a new chunk."""
        assert self.correlation_id == new_chunk.correlation_id
        return StreamingCodeExecutionResponseChunk(
            correlation_id=self.correlation_id, content=self.content + new_chunk.content
        )


class StreamingCodeExecutionResponse(RemoteExecutionResponse):
    namespace: ClassVar[Namespace] = "streaming_code_execution"

    content: str

    # Only present for shell code execution
    cancelled_for_timeout: bool = False
    exit_code: Optional[int] = None
    halted: bool = False


class CodeExecutionResponse(RemoteExecutionResponse):
    namespace: ClassVar[Namespace] = "code_execution"

    content: str

    # Only present for shell code execution
    cancelled_for_timeout: bool = False
    exit_code: Optional[int] = None
    halted: bool = False


class FileWriteResponse(RemoteExecutionResponse):
    namespace: ClassVar[Namespace] = "file_write"

    content: str


class ListFilesResponse(RemoteExecutionResponse):
    namespace: ClassVar[Namespace] = "list_files"

    files: list[RemoteFile]


class GetFileAttachmentResponse(RemoteExecutionResponse, FileAttachment):
    namespace: ClassVar[Namespace] = "get_file_attachment"

    exists: bool = Field(default=True)


class GetFileAttachmentsResponse(RemoteExecutionResponse):
    namespace: ClassVar[Namespace] = "get_file_attachments"

    file_attachments: list[FileAttachment]


class GetMatchingFilesResponse(RemoteExecutionResponse):
    namespace: ClassVar[Namespace] = "get_matching_files"

    files: list[RemoteFile]


class GetAllTrackedFilesResponse(RemoteExecutionResponse):
    namespace: ClassVar[Namespace] = "get_all_tracked_files"

    files: list[RemoteFile]


class SystemContextResponse(RemoteExecutionResponse):
    namespace: ClassVar[Namespace] = "system_context"

    exponent_txt: Optional[str]
    system_info: Optional[SystemInfo]


class HaltResponse(RemoteExecutionResponse):
    namespace: ClassVar[Namespace] = "halt"


class SwitchCLIChatResponse(RemoteExecutionResponse):
    namespace: ClassVar[Namespace] = "switch_cli_chat"


class ErrorResponse(RemoteExecutionResponse):
    namespace: ClassVar[Namespace] = "error"
    # The namespace of the request that caused the error.
    # Not a Namespace to avoid deserialization errors
    request_namespace: str
    error_type: ErrorType
    error_info: Optional[SerializableErrorInfo] = None

    @property
    def result_key(self) -> str:
        # Match the key of the request that caused the error
        return f"{self.request_namespace}:{self.correlation_id}"


class GitFileChange(BaseModel):
    path: str
    lines_added: int
    lines_deleted: int


class GitDiff(BaseModel):
    files: list[GitFileChange]
    truncated: bool = False  # True if there were more files than the limit
    total_files: int  # Total number of files changed, even if truncated


class GitCommitMetadata(BaseModel):
    author_name: str
    author_email: str
    author_date: str
    commit_date: str
    commit_message: str
    branch: str


class CreateCheckpointResponse(RemoteExecutionResponse):
    namespace: ClassVar[Namespace] = "create_checkpoint"

    correlation_id: str
    head_commit_hash: str
    head_commit_metadata: GitCommitMetadata
    uncommitted_changes_commit_hash: Optional[str] = None
    diff_versus_last_checkpoint: Optional[GitDiff] = None

    debug_info: Optional[dict[str, Any]] = None


class RollbackToCheckpointResponse(RemoteExecutionResponse):
    namespace: ClassVar[Namespace] = "rollback_to_checkpoint"

    debug_info: Optional[dict[str, Any]] = None


### Request Types


class RemoteExecutionRequestData(RemoteExecutionMessageData):
    pass


class RemoteExecutionRequest(RemoteExecutionMessage, Generic[ResponseT]):
    direction: ClassVar[Direction] = "request"


class CodeExecutionRequest(RemoteExecutionRequest[CodeExecutionResponse]):
    namespace: ClassVar[Namespace] = "code_execution"

    language: SupportedLanguage
    content: str
    timeout: int


class StreamingCodeExecutionRequest(
    RemoteExecutionRequest[
        Union[StreamingCodeExecutionResponseChunk, StreamingCodeExecutionResponse]
    ]
):
    namespace: ClassVar[Namespace] = "streaming_code_execution"

    language: SupportedLanguage
    content: str
    timeout: int


class HaltRequest(RemoteExecutionRequest[HaltResponse]):
    namespace: ClassVar[Namespace] = "halt"


class SwitchCLIChatRequest(RemoteExecutionRequest[SwitchCLIChatResponse]):
    namespace: ClassVar[Namespace] = "switch_cli_chat"
    new_chat_uuid: str


class FileWriteRequest(RemoteExecutionRequest[FileWriteResponse]):
    namespace: ClassVar[Namespace] = "file_write"

    file_path: str
    # Note we don't use SupportedLanguage here because we don't
    # require language-specific execution support for file writes
    language: str
    write_strategy: FileWriteStrategyName
    content: str


class ListFilesRequest(RemoteExecutionRequest[ListFilesResponse]):
    namespace: ClassVar[Namespace] = "list_files"

    directory: str


class GetFileAttachmentRequest(RemoteExecutionRequest[GetFileAttachmentResponse]):
    namespace: ClassVar[Namespace] = "get_file_attachment"

    file: RemoteFile


class GetFileAttachmentsRequest(RemoteExecutionRequest[GetFileAttachmentsResponse]):
    namespace: ClassVar[Namespace] = "get_file_attachments"

    files: list[RemoteFile]


class GetMatchingFilesRequest(RemoteExecutionRequest[GetMatchingFilesResponse]):
    namespace: ClassVar[Namespace] = "get_matching_files"

    search_term: str


class GetAllTrackedFilesRequest(RemoteExecutionRequest[GetAllTrackedFilesResponse]):
    namespace: ClassVar[Namespace] = "get_all_tracked_files"


class SystemContextRequest(RemoteExecutionRequest[SystemContextResponse]):
    namespace: ClassVar[Namespace] = "system_context"


class CreateCheckpointRequest(RemoteExecutionRequest[CreateCheckpointResponse]):
    namespace: ClassVar[Namespace] = "create_checkpoint"

    last_checkpoint_head_commit: Optional[str] = None
    last_checkpoint_uncommitted_changes_commit: Optional[str] = None


class RollbackToCheckpointRequest(RemoteExecutionRequest[RollbackToCheckpointResponse]):
    namespace: ClassVar[Namespace] = "rollback_to_checkpoint"

    head_commit: str
    uncommitted_changes_commit: Optional[str]


### Commands


### Command Response Types


class CommandResponse(RemoteExecutionResponse):
    namespace: ClassVar[Namespace] = "command"

    content: str
    content_json: dict[str, Any] = Field(default_factory=dict)
    subcommand: str = "unknown"


### Command Request Types


class CommandRequest(RemoteExecutionRequest[CommandResponse]):
    namespace: ClassVar[Namespace] = "command"

    data: CommandDataType = Field(..., discriminator="type")


RemoteExecutionRequestType = Union[
    CodeExecutionRequest,
    FileWriteRequest,
    ListFilesRequest,
    GetFileAttachmentRequest,
    GetFileAttachmentsRequest,
    GetMatchingFilesRequest,
    SystemContextRequest,
    GetAllTrackedFilesRequest,
    CommandRequest,
    HaltRequest,
    StreamingCodeExecutionRequest,
    SwitchCLIChatRequest,
    CreateCheckpointRequest,
    RollbackToCheckpointRequest,
]

RemoteExecutionResponseType = Union[
    CodeExecutionResponse,
    StreamingCodeExecutionResponseChunk,
    StreamingCodeExecutionResponse,
    FileWriteResponse,
    ListFilesResponse,
    GetFileAttachmentResponse,
    GetFileAttachmentsResponse,
    GetMatchingFilesResponse,
    GetAllTrackedFilesResponse,
    SystemContextResponse,
    CommandResponse,
    HaltResponse,
    SwitchCLIChatResponse,
    ErrorResponse,
    CreateCheckpointResponse,
    RollbackToCheckpointResponse,
]

StreamingResponseType = Union[
    StreamingCodeExecutionResponseChunk,
    StreamingCodeExecutionResponse,
    ErrorResponse,
]

STREAMING_NAMESPACES = [
    "streaming_code_execution",
    "streaming_code_execution_chunk",
]


class ChatMode(str, Enum):
    DEFAULT = "DEFAULT"  # chat just with model
    CLI = "CLI"
    CLOUD = "CLOUD"  # chat with cloud devbox
    PYTHON_INTERPRETER = "PYTHON_INTERPRETER"
    DATABASE = "DATABASE"  # chat with database connection
    WORKFLOW = "WORKFLOW"


class CLIConnectedState(BaseModel):
    chat_uuid: str
    connected: bool
    last_connected_at: Optional[datetime.datetime]
    connection_latency_ms: Optional[int]
    system_info: Optional[SystemInfo]
    exponent_version: Optional[str] = None
    editable_installation: bool = False


class DevboxConnectedState(str, Enum):
    # TODO: Only needed if we create devbox async
    INITIALIZED = "INITIALIZED"
    # The chat has been initialized, but the devbox is still loading
    DEVBOX_LOADING = "DEVBOX_LOADING"
    # CLI is connected and running on devbox
    CONNECTED = "CONNECTED"
    # CLI has disconnected
    # TODO: what condition?
    CLI_DISCONNECTED = "CLI_DISCONNECTED"
    # CLI has an error, devbox is running
    CLI_ERROR = "CLI_ERROR"
    # Devbox has an error
    DEVBOX_ERROR = "DEVBOX_ERROR"
    # Devbox_shutdown
    # TODO: In theory our terminal state, do we want to name something different?
    DEVBOX_SHUTDOWN = "DEVBOX_SHUTDOWN"


class CloudConnectedState(BaseModel):
    chat_uuid: str
    connected_state: DevboxConnectedState
    last_connected_at: Optional[datetime.datetime]
    system_info: Optional[SystemInfo]


class CLIErrorLog(BaseModel):
    event_data: str
    timestamp: datetime.datetime = datetime.datetime.now()
    attachment_data: Optional[str] = None
    version: Optional[str] = None
    chat_uuid: Optional[str] = None

    @property
    def loaded_event_data(self) -> Optional[Any]:
        try:
            return json.loads(self.event_data)
        except json.JSONDecodeError:
            return None

    @property
    def attachment_bytes(self) -> Optional[bytes]:
        if not self.attachment_data:
            return None
        return self.attachment_data.encode()



from typing import Generic, Literal, Optional, Protocol, TypeVar, Union
from exponent.core.types.command_data import CommandDataType
from pydantic import BaseModel, Field, JsonValue, ValidationInfo, field_validator


DEFAULT_CODE_BLOCK_TIMEOUT = 30

FileWriteStrategyName = Literal[
    "FULL_FILE_REWRITE", "UDIFF", "SEARCH_REPLACE", "NATURAL_EDIT"
]
WRITE_STRATEGY_NATURAL_EDIT: Literal["NATURAL_EDIT"] = "NATURAL_EDIT"
WRITE_STRATEGY_FULL_FILE_REWRITE: Literal["FULL_FILE_REWRITE"] = "FULL_FILE_REWRITE"
WRITE_STRATEGY_UDIFF: Literal["UDIFF"] = "UDIFF"
WRITE_STRATEGY_SEARCH_REPLACE: Literal["SEARCH_REPLACE"] = "SEARCH_REPLACE"


class ExponentEvent(BaseModel):
    chat_uuid: str
    event_uuid: str
    parent_uuid: Optional[str]
    turn_uuid: str

    metadata: dict[str, JsonValue] = Field(default_factory=dict)


class CodeBlockEvent(ExponentEvent):
    language: str
    content: str
    timeout: int = DEFAULT_CODE_BLOCK_TIMEOUT
    require_confirmation: bool = False


class EditContent(BaseModel):
    content: str
    original_file: Optional[str] = None


class NaturalEditContent(BaseModel):
    natural_edit: str
    intermediate_edit: Optional[str]
    original_file: Optional[str]
    new_file: Optional[str]
    error_content: Optional[str]

    @property
    def is_resolved(self) -> bool:
        return self.new_file is not None or self.error_content is not None

    @property
    def is_noop(self) -> bool:
        return bool(
            self.new_file is not None
            and self.original_file is not None
            and self.new_file == self.original_file
        )


class FileWriteEvent(ExponentEvent):
    file_path: str
    language: str
    write_strategy: FileWriteStrategyName
    write_content: Union[NaturalEditContent, EditContent]
    content: str
    error_content: Optional[str]
    require_confirmation: bool = False

    @field_validator("write_content")
    def validate_write_content_type(
        cls, v: Union[NaturalEditContent, EditContent], info: ValidationInfo
    ) -> Union[NaturalEditContent, EditContent]:
        write_strategy = info.data.get("write_strategy")
        if write_strategy == WRITE_STRATEGY_NATURAL_EDIT:
            if not isinstance(v, NaturalEditContent):
                raise ValueError(
                    "When write_strategy is NATURAL_EDIT, write_content must be NaturalEditContent"
                )
        else:
            if not isinstance(v, EditContent):
                raise ValueError(
                    "For non-NATURAL_EDIT strategies, write_content must be EditContent"
                )
        return v


T = TypeVar("T", bound=CommandDataType)


class HoldsCommandData(Protocol, Generic[T]):
    data: T


class CommandEvent(ExponentEvent):
    data: CommandDataType = Field(..., discriminator="type")
    require_confirmation: bool = False


LocalEventType = Union[FileWriteEvent, CodeBlockEvent, CommandEvent]



from __future__ import annotations

import asyncio
import json
import logging
from collections.abc import AsyncGenerator, Callable
from contextlib import asynccontextmanager
from typing import Optional, TypeVar, Union, Dict

from dataclasses import dataclass
from exponent.core.config import is_editable_install
from exponent.core.remote_execution.error_info import serialize_error_info
from exponent.core.types.event_types import (
    CodeBlockEvent,
    CommandEvent,
    FileWriteEvent,
    LocalEventType,
)
from exponent.utils.version import get_installed_version
import websockets.client
import websockets.exceptions
from httpx import (
    AsyncClient,
    codes as http_status,
)
from pydantic import BaseModel

from exponent.commands.utils import ConnectionTracker
from exponent.core.remote_execution import files, system_context
from exponent.core.remote_execution.code_execution import (
    execute_code,
    execute_code_streaming,
)
from exponent.core.remote_execution.command_execution import execute_command
from exponent.core.remote_execution.file_write import execute_file_write
from exponent.core.remote_execution.session import (
    RemoteExecutionClientSession,
    get_session,
)
from exponent.core.remote_execution.checkpoints import (
    create_checkpoint,
    rollback_to_checkpoint,
)
from exponent.core.remote_execution.types import (
    CLIConnectedState,
    CodeExecutionRequest,
    StreamingCodeExecutionRequest,
    CommandRequest,
    CreateChatResponse,
    ExecutionEndResponse,
    FileWriteRequest,
    GetAllTrackedFilesRequest,
    GetFileAttachmentRequest,
    GetFileAttachmentsRequest,
    GetMatchingFilesRequest,
    HaltRequest,
    HaltResponse,
    HeartbeatInfo,
    ListFilesRequest,
    RemoteExecutionRequestType,
    RemoteExecutionResponseType,
    RunWorkflowRequest,
    StartChatRequest,
    StartChatResponse,
    SwitchCLIChatRequest,
    SwitchCLIChatResponse,
    SystemContextRequest,
    UseToolsConfig,
    ErrorResponse,
    CreateCheckpointRequest,
    RollbackToCheckpointRequest,
)
from exponent.core.remote_execution.utils import (
    convert_event_to_execution_request,
    deserialize_api_response,
    deserialize_request_data,
    serialize_message,
)

logger = logging.getLogger(__name__)


TModel = TypeVar("TModel", bound=BaseModel)


@dataclass
class WSDisconnected:
    error_message: Optional[str] = None


@dataclass
class SwitchCLIChat:
    new_chat_uuid: str


REMOTE_EXECUTION_CLIENT_EXIT_INFO = Union[WSDisconnected, SwitchCLIChat]


class RemoteExecutionClient:
    def __init__(
        self,
        session: RemoteExecutionClientSession,
    ):
        self.current_session = session

        self.file_cache: files.FileCache = files.FileCache(session.working_directory)

        # for active code executions, track whether they should be halted
        # correlation_id -> should_halt
        self._halt_states: Dict[str, bool] = {}
        self._halt_lock = asyncio.Lock()

    @property
    def working_directory(self) -> str:
        return self.current_session.working_directory

    @property
    def api_client(self) -> AsyncClient:
        return self.current_session.api_client

    @property
    def ws_client(self) -> AsyncClient:
        return self.current_session.ws_client

    async def add_code_execution_to_halt_states(self, correlation_id: str) -> None:
        async with self._halt_lock:
            self._halt_states[correlation_id] = False

    async def halt_all_code_executions(self) -> None:
        async with self._halt_lock:
            self._halt_states = {
                correlation_id: True for correlation_id in self._halt_states.keys()
            }

    async def clear_halt_state(self, correlation_id: str) -> None:
        async with self._halt_lock:
            self._halt_states.pop(correlation_id, None)

    def get_halt_check(self, correlation_id: str) -> Callable[[], bool]:
        def should_halt() -> bool:
            # Don't need to lock here, since just reading from dict
            return self._halt_states.get(correlation_id, False)

        return should_halt

    async def _handle_websocket_message(
        self,
        msg: str,
        websocket: websockets.client.WebSocketClientProtocol,
        requests: asyncio.Queue[RemoteExecutionRequestType],
    ) -> Optional[REMOTE_EXECUTION_CLIENT_EXIT_INFO]:
        """Handle an incoming websocket message.
        Returns None to continue processing, or a REMOTE_EXECUTION_CLIENT_EXIT_INFO to exit."""

        msg_data = json.loads(msg)
        if msg_data["type"] != "request":
            return None

        data = json.dumps(msg_data["data"])
        request = deserialize_request_data(data)

        if isinstance(request, HaltRequest):
            logger.debug(
                f"Received HaltRequest for correlation_id {request.correlation_id}"
            )
            await self.halt_all_code_executions()
            data = json.loads(
                serialize_message(
                    HaltResponse(
                        correlation_id=request.correlation_id,
                    )
                )
            )
            msg = json.dumps({"type": "result", "data": data})
            logger.debug(
                f"Sending HaltResponse for correlation_id {request.correlation_id}"
            )
            await websocket.send(msg)
            return None

        elif isinstance(request, SwitchCLIChatRequest):
            data = json.loads(
                serialize_message(
                    SwitchCLIChatResponse(
                        correlation_id=request.correlation_id,
                    )
                )
            )
            msg = json.dumps({"type": "result", "data": data})
            await websocket.send(msg)
            return SwitchCLIChat(new_chat_uuid=request.new_chat_uuid)

        elif isinstance(request, (CodeExecutionRequest, StreamingCodeExecutionRequest)):
            await self.add_code_execution_to_halt_states(request.correlation_id)
            await requests.put(request)
            return None

        else:
            await requests.put(request)
            return None

    async def _setup_tasks(
        self,
        beats: asyncio.Queue[HeartbeatInfo],
        requests: asyncio.Queue[RemoteExecutionRequestType],
        results: asyncio.Queue[RemoteExecutionResponseType],
    ) -> list[asyncio.Task[None]]:
        """Setup heartbeat and executor tasks."""

        async def beat() -> None:
            while True:
                info = self.get_heartbeat_info()
                await beats.put(info)
                await asyncio.sleep(3)

        # Lock to ensure that only one executor can grab a
        # request at a time.
        requests_lock = asyncio.Lock()

        # Lock to ensure that only one executor can put a
        # result in the results queue at a time.
        results_lock = asyncio.Lock()

        async def executor() -> None:
            # We use locks here to protect the request/result
            # queues from being accessed by multiple executors.
            while True:
                async with requests_lock:
                    request = await requests.get()

                try:
                    if isinstance(request, StreamingCodeExecutionRequest):
                        async for streaming_response in self.handle_streaming_request(
                            request
                        ):
                            async with results_lock:
                                await results.put(streaming_response)
                    else:
                        # Note that we don't want to hold the lock here
                        # because we want other executors to be able to
                        # grab requests while we're handling a request.
                        response = await self.handle_request(request)
                        async with results_lock:
                            await results.put(response)
                except Exception as e:
                    logger.info(f"Error handling request {request.namespace}:\n\n{e}")
                    async with results_lock:
                        await results.put(
                            ErrorResponse(
                                error_type="request_error",
                                correlation_id=request.correlation_id,
                                request_namespace=request.namespace,
                                error_info=serialize_error_info(e),
                            )
                        )

        beat_task = asyncio.create_task(beat())
        # Three parallel executors to handle requests

        executor_tasks = [
            asyncio.create_task(executor()),
            asyncio.create_task(executor()),
            asyncio.create_task(executor()),
        ]

        return [beat_task] + executor_tasks

    async def _process_websocket_messages(
        self,
        websocket: websockets.client.WebSocketClientProtocol,
        beats: asyncio.Queue[HeartbeatInfo],
        requests: asyncio.Queue[RemoteExecutionRequestType],
        results: asyncio.Queue[RemoteExecutionResponseType],
    ) -> REMOTE_EXECUTION_CLIENT_EXIT_INFO:
        """Process messages from the websocket connection."""
        try:
            recv = asyncio.create_task(websocket.recv())
            get_beat = asyncio.create_task(beats.get())
            get_result = asyncio.create_task(results.get())
            pending = {recv, get_beat, get_result}

            while True:
                done, pending = await asyncio.wait(
                    pending, return_when=asyncio.FIRST_COMPLETED
                )

                if recv in done:
                    msg = str(recv.result())
                    exit_info = await self._handle_websocket_message(
                        msg, websocket, requests
                    )
                    if exit_info is not None:
                        return exit_info

                    recv = asyncio.create_task(websocket.recv())
                    pending.add(recv)

                if get_beat in done:
                    info = get_beat.result()
                    data = json.loads(info.model_dump_json())
                    msg = json.dumps({"type": "heartbeat", "data": data})
                    await websocket.send(msg)

                    get_beat = asyncio.create_task(beats.get())
                    pending.add(get_beat)

                if get_result in done:
                    response = get_result.result()
                    data = json.loads(serialize_message(response))
                    msg = json.dumps({"type": "result", "data": data})
                    await websocket.send(msg)

                    get_result = asyncio.create_task(results.get())
                    pending.add(get_result)
        finally:
            for task in pending:
                task.cancel()

            await asyncio.gather(*pending, return_exceptions=True)

    async def _handle_websocket_connection(
        self,
        websocket: websockets.client.WebSocketClientProtocol,
        connection_tracker: Optional[ConnectionTracker],
    ) -> Optional[REMOTE_EXECUTION_CLIENT_EXIT_INFO]:
        """Handle a single websocket connection.
        Returns None to continue with reconnection attempts, or an exit info to terminate."""
        if connection_tracker is not None:
            await connection_tracker.set_connected(True)

        beats: asyncio.Queue[HeartbeatInfo] = asyncio.Queue()
        requests: asyncio.Queue[RemoteExecutionRequestType] = asyncio.Queue()
        results: asyncio.Queue[RemoteExecutionResponseType] = asyncio.Queue()

        tasks = await self._setup_tasks(beats, requests, results)

        try:
            return await self._process_websocket_messages(
                websocket, beats, requests, results
            )
        except websockets.exceptions.ConnectionClosed as e:
            if e.rcvd is not None:
                if e.rcvd.code == 1000:
                    # Normal closure, exit completely
                    return WSDisconnected()
                elif e.rcvd.code == 1008:
                    error_message = (
                        "Error connecting to websocket"
                        if e.rcvd.reason is None
                        else e.rcvd.reason
                    )
                    return WSDisconnected(error_message=error_message)
            # Otherwise, allow reconnection attempt
            return None
        except TimeoutError:
            # Timeout, allow reconnection attempt
            # TODO: investgate if this is needed, possibly scope it down
            return None
        finally:
            for task in tasks:
                task.cancel()
            await asyncio.gather(*tasks, return_exceptions=True)
            if connection_tracker is not None:
                await connection_tracker.set_connected(False)

    async def run_connection(
        self,
        chat_uuid: str,
        connection_tracker: Optional[ConnectionTracker] = None,
    ) -> REMOTE_EXECUTION_CLIENT_EXIT_INFO:
        """Run the websocket connection loop."""
        self.current_session.set_chat_uuid(chat_uuid)

        async for websocket in self.ws_connect(f"/api/ws/chat/{chat_uuid}"):
            result = await self._handle_websocket_connection(
                websocket, connection_tracker
            )
            if result is not None:
                return result
            # If we get None, we'll try to reconnect

        # If we exit the websocket connection loop without returning,
        # it means we couldn't establish a connection
        return WSDisconnected(error_message="Could not establish websocket connection")

    async def check_remote_end_event(self, chat_uuid: str) -> bool:
        response = await self.api_client.get(
            f"/api/remote_execution/{chat_uuid}/execution_end",
        )
        execution_end_response = await deserialize_api_response(
            response, ExecutionEndResponse
        )
        return execution_end_response.execution_ended

    async def create_chat(self) -> CreateChatResponse:
        response = await self.api_client.post(
            "/api/remote_execution/create_chat",
        )
        return await deserialize_api_response(response, CreateChatResponse)

    async def start_chat(
        self, chat_uuid: str, prompt: str, use_tools_config: UseToolsConfig
    ) -> StartChatResponse:
        response = await self.api_client.post(
            "/api/remote_execution/start_chat",
            json=StartChatRequest(
                chat_uuid=chat_uuid,
                prompt=prompt,
                use_tools_config=use_tools_config,
            ).model_dump(),
            timeout=60,
        )
        return await deserialize_api_response(response, StartChatResponse)

    async def run_workflow(self, chat_uuid: str, workflow_id: str) -> None:
        response = await self.api_client.post(
            "/api/remote_execution/run_workflow",
            json=RunWorkflowRequest(
                chat_uuid=chat_uuid,
                workflow_id=workflow_id,
            ).model_dump(),
            timeout=60,
        )
        if response.status_code != http_status.OK:
            raise Exception(
                f"Failed to run workflow with status code {response.status_code} and response {response.text}"
            )

    def get_heartbeat_info(self) -> HeartbeatInfo:
        return HeartbeatInfo(
            system_info=system_context.get_system_info(self.working_directory),
            exponent_version=get_installed_version(),
            editable_installation=is_editable_install(),
        )

    async def send_heartbeat(self, chat_uuid: str) -> CLIConnectedState:
        logger.info(f"Sending heartbeat for chat_uuid {chat_uuid}")
        heartbeat_info = self.get_heartbeat_info()
        response = await self.api_client.post(
            f"/api/remote_execution/{chat_uuid}/heartbeat",
            content=heartbeat_info.model_dump_json(),
            timeout=60,
        )
        if response.status_code != http_status.OK:
            raise Exception(
                f"Heartbeat failed with status code {response.status_code} and response {response.text}"
            )
        connected_state = await deserialize_api_response(response, CLIConnectedState)
        logger.info(f"Heartbeat response: {connected_state}")
        return connected_state

    async def handle_request(
        self, request: Union[RemoteExecutionRequestType, LocalEventType]
    ) -> RemoteExecutionResponseType:
        if isinstance(request, (CodeBlockEvent, FileWriteEvent, CommandEvent)):
            request = convert_event_to_execution_request(request)

        try:
            response: RemoteExecutionResponseType
            if isinstance(request, CodeExecutionRequest):
                response = await execute_code(
                    request,
                    self.current_session,
                    working_directory=self.working_directory,
                    should_halt=self.get_halt_check(request.correlation_id),
                )
            elif isinstance(request, FileWriteRequest):
                response = execute_file_write(
                    request,
                    working_directory=self.working_directory,
                )
            elif isinstance(request, ListFilesRequest):
                response = await files.list_files(request)
            elif isinstance(request, GetFileAttachmentRequest):
                response = await files.get_file_attachment(
                    request, self.working_directory
                )
            elif isinstance(request, GetFileAttachmentsRequest):
                response = await files.get_file_attachments(
                    request, self.working_directory
                )
            elif isinstance(request, GetMatchingFilesRequest):
                response = await files.get_matching_files(request, self.file_cache)
            elif isinstance(request, SystemContextRequest):
                response = system_context.get_system_context(
                    request, self.working_directory
                )
            elif isinstance(request, GetAllTrackedFilesRequest):
                response = await files.get_all_tracked_files(
                    request, self.working_directory
                )
            elif isinstance(request, CommandRequest):
                response = await execute_command(
                    request,
                    self.working_directory,
                )
            elif isinstance(request, StreamingCodeExecutionRequest):
                assert False, "CodeExecutionStreamingRequest should be sent to handle_streaming_request"
            elif isinstance(request, HaltRequest):
                assert False, "HaltRequest should be handled in the main loop"
            elif isinstance(request, SwitchCLIChatRequest):
                assert False, "SwitchCLIChatRequest should be handled in the main loop"
            elif isinstance(request, CreateCheckpointRequest):
                response = await create_checkpoint(request)
            elif isinstance(request, RollbackToCheckpointRequest):
                response = await rollback_to_checkpoint(
                    request.correlation_id,
                    request.head_commit,
                    request.uncommitted_changes_commit,
                )
            else:
                logger.info(f"Unhandled request type: {type(request)}")
                response = ErrorResponse(
                    error_type="unknown_request_type",
                    correlation_id=request.correlation_id,
                    request_namespace=request.namespace,
                )
            return response
        finally:
            # Clean up halt state after request is complete
            if isinstance(request, CodeExecutionRequest) or isinstance(
                request, StreamingCodeExecutionRequest
            ):
                await self.clear_halt_state(request.correlation_id)

    async def handle_streaming_request(
        self, request: StreamingCodeExecutionRequest
    ) -> AsyncGenerator[RemoteExecutionResponseType, None]:
        if not isinstance(request, StreamingCodeExecutionRequest):
            assert False, f"{type(request)} should be sent to handle_streaming_request"

        async for output in execute_code_streaming(
            request,
            self.current_session,
            working_directory=self.working_directory,
            should_halt=self.get_halt_check(request.correlation_id),
        ):
            yield output

    def ws_connect(self, path: str) -> websockets.client.connect:
        base_url = (
            str(self.ws_client.base_url)
            .replace("http://", "ws://")
            .replace("https://", "wss://")
        )

        url = f"{base_url}{path}"
        headers = {"api-key": self.api_client.headers["api-key"]}

        conn = websockets.client.connect(
            url, extra_headers=headers, timeout=10, ping_timeout=10
        )

        # Stop exponential backoff from blowing up
        # the wait time between connection attempts
        conn.BACKOFF_MAX = 2
        conn.BACKOFF_INITIAL = 1  # pyright: ignore

        return conn

    @staticmethod
    @asynccontextmanager
    async def session(
        api_key: str, base_url: str, base_ws_url: str, working_directory: str
    ) -> AsyncGenerator[RemoteExecutionClient, None]:
        async with get_session(
            working_directory, base_url, base_ws_url, api_key
        ) as session:
            yield RemoteExecutionClient(session)



# Auto-generated, do not edit directly. Run `make generate_strategy_info` to update.

import enum

from pydantic import BaseModel

from exponent.core.remote_execution.types import ChatMode


class StrategyName(str, enum.Enum):
    NATURAL_EDIT = "NATURAL_EDIT"
    NATURAL_EDIT_CLAUDE_3_7_XML = "NATURAL_EDIT_CLAUDE_3_7_XML"
    NATURAL_EDIT_CLAUDE_3_7_XML_WITH_STEPS = "NATURAL_EDIT_CLAUDE_3_7_XML_WITH_STEPS"
    NATURAL_EDIT_CLAUDE_3_7_FUNCTION_CALLING = (
        "NATURAL_EDIT_CLAUDE_3_7_FUNCTION_CALLING"
    )
    NATURAL_EDIT_EXPERIMENTAL = "NATURAL_EDIT_EXPERIMENTAL"
    FULL_FILE_REWRITE = "FULL_FILE_REWRITE"
    STREAMING_NATURAL_EDIT = "STREAMING_NATURAL_EDIT"
    FUNCTION_CALLING = "FUNCTION_CALLING"
    RAW_GPT = "RAW_GPT"
    STRICT_NATURAL_EDIT = "STRICT_NATURAL_EDIT"
    AGENT = "AGENT"
    AGENT_STEP = "AGENT_STEP"
    DATABASE = "DATABASE"
    DATABASE_AGENT = "DATABASE_AGENT"
    DATABASE_AGENT_STEP = "DATABASE_AGENT_STEP"
    REVIEW_FILES = "REVIEW_FILES"
    EXPLORE_CODEBASE = "EXPLORE_CODEBASE"
    GENERATE_REVIEW = "GENERATE_REVIEW"


class StrategyInfo(BaseModel):
    strategy_name: StrategyName
    display_name: str
    description: str
    disabled: bool
    display_order: int
    is_agentic: bool


CHAT_MODE_DEFAULTS: dict[ChatMode, StrategyName] = {
    ChatMode.DEFAULT: StrategyName.RAW_GPT,
    ChatMode.CLI: StrategyName.NATURAL_EDIT_CLAUDE_3_7_XML,
    ChatMode.CLOUD: StrategyName.NATURAL_EDIT_CLAUDE_3_7_XML,
    ChatMode.DATABASE: StrategyName.DATABASE,
    ChatMode.PYTHON_INTERPRETER: StrategyName.NATURAL_EDIT_CLAUDE_3_7_XML,
    ChatMode.WORKFLOW: StrategyName.NATURAL_EDIT,
}

STRATEGY_INFO_LIST: list[StrategyInfo] = [
    StrategyInfo(
        strategy_name=StrategyName.NATURAL_EDIT,
        display_name="Natural Edit",
        description="A natural file editing strategy.",
        disabled=False,
        display_order=0,
        is_agentic=False,
    ),
    StrategyInfo(
        strategy_name=StrategyName.NATURAL_EDIT_CLAUDE_3_7_XML,
        display_name="Natural Edit Claude 3.7 XML",
        description="A natural file editing strategy that uses Claude 3.7 XML.",
        disabled=True,
        display_order=0,
        is_agentic=False,
    ),
    StrategyInfo(
        strategy_name=StrategyName.NATURAL_EDIT_CLAUDE_3_7_XML_WITH_STEPS,
        display_name="Natural Edit Claude 3.7 XML with Steps",
        description="A natural file editing strategy that uses Claude 3.7 XML with steps.",
        disabled=True,
        display_order=0,
        is_agentic=False,
    ),
    StrategyInfo(
        strategy_name=StrategyName.NATURAL_EDIT_CLAUDE_3_7_FUNCTION_CALLING,
        display_name="Natural Edit Claude 3.7 Function Calling",
        description="A natural file editing strategy that uses Claude 3.7 Function Calling.",
        disabled=True,
        display_order=0,
        is_agentic=True,
    ),
    StrategyInfo(
        strategy_name=StrategyName.NATURAL_EDIT_EXPERIMENTAL,
        display_name="Natural Edit Experimental",
        description="A natural file editing strategy.",
        disabled=True,
        display_order=0,
        is_agentic=False,
    ),
    StrategyInfo(
        strategy_name=StrategyName.STREAMING_NATURAL_EDIT,
        display_name="Streaming Natural Edit",
        description="A natural file editing strategy that streams edits.",
        disabled=True,
        display_order=1,
        is_agentic=False,
    ),
    StrategyInfo(
        strategy_name=StrategyName.FULL_FILE_REWRITE,
        display_name="Full File Rewrites",
        description="Rewrites the full file every time. Use this if your files are generally less than 300 lines.",
        disabled=False,
        display_order=2,
        is_agentic=False,
    ),
    StrategyInfo(
        strategy_name=StrategyName.FUNCTION_CALLING,
        display_name="Function Calling",
        description="No description",
        disabled=True,
        display_order=99,
        is_agentic=False,
    ),
    StrategyInfo(
        strategy_name=StrategyName.RAW_GPT,
        display_name="Raw GPT",
        description="No description",
        disabled=True,
        display_order=99,
        is_agentic=False,
    ),
    StrategyInfo(
        strategy_name=StrategyName.STRICT_NATURAL_EDIT,
        display_name="Natural Edit No FFR",
        description="A natural file editing strategy that does not use full file rewrites.",
        disabled=True,
        display_order=99,
        is_agentic=False,
    ),
    StrategyInfo(
        strategy_name=StrategyName.AGENT,
        display_name="Agent",
        description="No description",
        disabled=True,
        display_order=99,
        is_agentic=True,
    ),
    StrategyInfo(
        strategy_name=StrategyName.AGENT_STEP,
        display_name="Agent Step",
        description="No description",
        disabled=True,
        display_order=99,
        is_agentic=False,
    ),
    StrategyInfo(
        strategy_name=StrategyName.DATABASE,
        display_name="Database",
        description="No description",
        disabled=True,
        display_order=99,
        is_agentic=False,
    ),
    StrategyInfo(
        strategy_name=StrategyName.DATABASE_AGENT,
        display_name="Database Agent",
        description="No description",
        disabled=True,
        display_order=99,
        is_agentic=True,
    ),
    StrategyInfo(
        strategy_name=StrategyName.DATABASE_AGENT_STEP,
        display_name="Database Agent Step",
        description="No description",
        disabled=True,
        display_order=99,
        is_agentic=False,
    ),
    StrategyInfo(
        strategy_name=StrategyName.REVIEW_FILES,
        display_name="Review Files",
        description="No description",
        disabled=True,
        display_order=99,
        is_agentic=False,
    ),
    StrategyInfo(
        strategy_name=StrategyName.EXPLORE_CODEBASE,
        display_name="Explore Codebase",
        description="No description",
        disabled=True,
        display_order=99,
        is_agentic=False,
    ),
    StrategyInfo(
        strategy_name=StrategyName.GENERATE_REVIEW,
        display_name="Generate Review",
        description="No description",
        disabled=True,
        display_order=99,
        is_agentic=False,
    ),
]


ENABLED_STRATEGY_INFO_LIST: list[StrategyInfo] = [
    StrategyInfo(
        strategy_name=StrategyName.NATURAL_EDIT,
        display_name="Natural Edit",
        description="A natural file editing strategy.",
        disabled=False,
        display_order=0,
        is_agentic=False,
    ),
    StrategyInfo(
        strategy_name=StrategyName.FULL_FILE_REWRITE,
        display_name="Full File Rewrites",
        description="Rewrites the full file every time. Use this if your files are generally less than 300 lines.",
        disabled=False,
        display_order=2,
        is_agentic=False,
    ),
]



import json
import sys
from typing import Optional

import click

from exponent.commands.common import (
    redirect_to_login,
    run_until_complete,
    set_login_complete,
    refresh_api_key_task,
)
from exponent.commands.settings import use_settings
from exponent.commands.types import exponent_cli_group
from exponent.core.config import Settings, get_settings
from exponent.core.graphql.client import GraphQLClient
from exponent.core.graphql.cloud_config_queries import (
    CREATE_CLOUD_CONFIG_MUTATION,
    GET_CLOUD_CONFIGS_QUERY,
    UPDATE_CLOUD_CONFIG_MUTATION,
)
from exponent.core.graphql.get_chats_query import GET_CHATS_QUERY
from exponent.core.graphql.github_config_queries import (
    CREATE_GITHUB_CONFIG_MUTATION,
    CHECK_GITHUB_CONFIG_VALIDITY_QUERY,
    REPOS_FOR_GITHUB_CONFIG_QUERY,
)
from exponent.core.graphql.mutations import (
    UPGRADE_USER_TO_TRUSTED_MUTATION,
)
from exponent.core.graphql.subscriptions import AUTHENTICATED_USER_SUBSCRIPTION
from exponent.utils.version import (
    get_installed_metadata,
    get_installed_version,
    get_installer,
    get_python_path,
    get_sys_executable,
)


@exponent_cli_group()
def config_cli() -> None:
    """Manage Exponent configuration settings."""
    pass


@config_cli.command()
@use_settings
def debug(
    settings: Settings,
) -> None:
    click.echo("Settings:")
    click.echo(settings)
    click.echo("\nInstalled Version:")
    click.echo(get_installed_version())
    click.echo("\nPython Version:")
    click.echo(sys.version)
    click.echo("\nPython Path:")
    click.echo(get_python_path())
    click.echo("\nSys Executable Path:")
    click.echo(get_sys_executable())
    click.echo("\nInstaller:")
    click.echo(get_installer())
    click.echo("\nInstalled Metadata:")
    click.echo(get_installed_metadata())


@config_cli.command()
@use_settings
def check_github_config_validity(
    settings: Settings,
) -> None:
    if not settings.api_key:
        redirect_to_login(settings)
        return

    run_until_complete(
        check_github_config_validity_task(
            api_key=settings.api_key,
            base_api_url=settings.base_api_url,
            base_ws_url=settings.base_ws_url,
        )
    )


async def check_github_config_validity_task(
    api_key: str,
    base_api_url: str,
    base_ws_url: str,
) -> None:
    graphql_client = GraphQLClient(api_key, base_api_url, base_ws_url)
    result = await graphql_client.execute(CHECK_GITHUB_CONFIG_VALIDITY_QUERY)
    click.echo(result)


@config_cli.command()
@use_settings
def repos_for_github_config(
    settings: Settings,
) -> None:
    if not settings.api_key:
        redirect_to_login(settings)
        return

    run_until_complete(
        repos_for_github_config_task(
            api_key=settings.api_key,
            base_api_url=settings.base_api_url,
            base_ws_url=settings.base_ws_url,
        )
    )


async def repos_for_github_config_task(
    api_key: str,
    base_api_url: str,
    base_ws_url: str,
) -> None:
    graphql_client = GraphQLClient(api_key, base_api_url, base_ws_url)
    try:
        click.echo("Sending request to fetch repos...")
        result = await graphql_client.execute(
            REPOS_FOR_GITHUB_CONFIG_QUERY, timeout=120
        )  # 120 seconds timeout
        click.echo("Request completed. Result:")
        click.echo(result)
    except Exception as e:
        click.echo(f"An error occurred while fetching repos: {str(e)}")
        click.echo(f"Error type: {type(e).__name__}")
        # Add more detailed error information if available
        if hasattr(e, "response"):
            click.echo(f"Response status: {e.response.status_code}")
            click.echo(f"Response content: {e.response.text}")


@config_cli.command()
@click.option(
    "--set-git-warning-disabled",
    is_flag=True,
    help="Disable the git warning for running Exponent from a non-git repository",
)
@click.option(
    "--set-git-warning-enabled",
    is_flag=True,
    help="Enable the git warning for running Exponent from a non-git repository",
)
def config(set_git_warning_disabled: bool, set_git_warning_enabled: bool) -> None:
    """Display current Exponent configuration."""
    config_file_settings = get_settings().get_config_file_settings()

    if set_git_warning_enabled and set_git_warning_disabled:
        click.secho(
            "Cannot enable and disable git warning at the same time.",
            fg="red",
        )
        return

    elif set_git_warning_enabled:
        if (
            "options" in config_file_settings
            and "git_warning_disabled" in config_file_settings["options"]
        ):
            del config_file_settings["options"]["git_warning_disabled"]

        click.secho(
            "Git warning enabled. Exponent will now check for a git repository.\n",
            fg="yellow",
        )

    elif set_git_warning_disabled:
        if "options" not in config_file_settings:
            config_file_settings["options"] = {}

        config_file_settings["options"]["git_warning_disabled"] = True

        click.secho(
            "Git warning disabled. Exponent will no longer check for a git repository.\n",
            fg="yellow",
        )
    else:
        click.secho(
            json.dumps(config_file_settings, indent=2),
            fg="green",
        )


@config_cli.command()
@click.option("--key", help="Your Exponent API Key")
@use_settings
def login(settings: Settings, key: str) -> None:
    """Log in to Exponent."""

    if not key:
        redirect_to_login(settings, "provided")
        return

    click.echo("Verifying API key...")
    run_until_complete(
        set_login_complete(key, settings.base_api_url, settings.base_ws_url)
    )
    click.secho("Success!", fg="green")

    click.echo(f"Saving API Key to {settings.config_file_path}")

    if settings.api_key and settings.api_key != key:
        click.confirm("Detected existing API Key, continue? ", default=True, abort=True)

    settings.update_api_key(key)
    settings.write_settings_to_config_file()

    click.echo("API Key saved.")


@config_cli.command(hidden=True)
@use_settings
def get_chats(
    settings: Settings,
) -> None:
    if not settings.api_key:
        redirect_to_login(settings)
        return

    run_until_complete(
        get_chats_task(
            api_key=settings.api_key,
            base_api_url=settings.base_api_url,
            base_ws_url=settings.base_ws_url,
        )
    )


@config_cli.command(hidden=True)
@use_settings
def get_authenticated_user(
    settings: Settings,
) -> None:
    if not settings.api_key:
        redirect_to_login(settings)
        return

    run_until_complete(
        get_authenticated_user_task(
            api_key=settings.api_key,
            base_api_url=settings.base_api_url,
            base_ws_url=settings.base_ws_url,
        )
    )


async def get_chats_task(
    api_key: str,
    base_api_url: str,
    base_ws_url: str,
) -> None:
    graphql_client = GraphQLClient(api_key, base_api_url, base_ws_url)
    result = await graphql_client.execute(GET_CHATS_QUERY)
    click.echo(result)


async def get_authenticated_user_task(
    api_key: str,
    base_api_url: str,
    base_ws_url: str,
) -> None:
    graphql_client = GraphQLClient(api_key, base_api_url, base_ws_url)
    async for it in graphql_client.subscribe(AUTHENTICATED_USER_SUBSCRIPTION):
        click.echo(it)


@config_cli.command(hidden=True)
@use_settings
def get_cloud_configs(
    settings: Settings,
) -> None:
    if not settings.api_key:
        redirect_to_login(settings)
        return

    run_until_complete(
        get_cloud_configs_task(
            api_key=settings.api_key,
            base_api_url=settings.base_api_url,
            base_ws_url=settings.base_ws_url,
        )
    )


async def get_cloud_configs_task(
    api_key: str,
    base_api_url: str,
    base_ws_url: str,
) -> None:
    graphql_client = GraphQLClient(api_key, base_api_url, base_ws_url)
    result = await graphql_client.execute(GET_CLOUD_CONFIGS_QUERY)
    click.echo(result)


async def create_github_config_task(
    api_key: str,
    base_api_url: str,
    base_ws_url: str,
    github_pat: str,
) -> None:
    graphql_client = GraphQLClient(api_key, base_api_url, base_ws_url)
    variables = {
        "githubPat": github_pat,
    }
    result = await graphql_client.execute(CREATE_GITHUB_CONFIG_MUTATION, variables)
    click.echo(result)


@config_cli.command(hidden=True)
@click.option("--github-pat", required=True, help="Github personal access token")
@use_settings
def create_github_config(
    settings: Settings,
    github_pat: str,
) -> None:
    if not settings.api_key:
        redirect_to_login(settings)
        return

    run_until_complete(
        create_github_config_task(
            api_key=settings.api_key,
            base_api_url=settings.base_api_url,
            base_ws_url=settings.base_ws_url,
            github_pat=github_pat,
        )
    )


@config_cli.command(hidden=True)
@click.option("--github-org-name", required=True, help="GitHub organization name")
@click.option("--github-repo-name", required=True, help="GitHub repository name")
@click.option(
    "--setup_commands",
    required=False,
    help="List of commands to set up and build your repo",
    multiple=True,
)
@use_settings
def create_cloud_config(
    settings: Settings,
    github_org_name: str,
    github_repo_name: str,
    setup_commands: Optional[list[str]] = None,
) -> None:
    if not settings.api_key:
        redirect_to_login(settings)
        return

    run_until_complete(
        create_cloud_config_task(
            api_key=settings.api_key,
            base_api_url=settings.base_api_url,
            base_ws_url=settings.base_ws_url,
            github_org_name=github_org_name,
            github_repo_name=github_repo_name,
            setup_commands=setup_commands,
        )
    )


async def create_cloud_config_task(
    api_key: str,
    base_api_url: str,
    base_ws_url: str,
    github_org_name: str,
    github_repo_name: str,
    setup_commands: Optional[list[str]],
) -> None:
    graphql_client = GraphQLClient(api_key, base_api_url, base_ws_url)
    variables = {
        "githubOrgName": github_org_name,
        "githubRepoName": github_repo_name,
        "setupCommands": setup_commands,
    }
    result = await graphql_client.execute(CREATE_CLOUD_CONFIG_MUTATION, variables)
    click.echo(result)


@config_cli.command(hidden=True)
@click.option("--cloud-config-uuid", required=True, help="Cloud config UUID")
@click.option("--github-org-name", required=True, help="GitHub organization name")
@click.option("--github-repo-name", required=True, help="GitHub repository name")
@click.option(
    "--setup_commands",
    required=False,
    help="List of commands to set up and build your repo",
    multiple=True,
)
@use_settings
def update_cloud_config(
    settings: Settings,
    cloud_config_uuid: str,
    github_org_name: str,
    github_repo_name: str,
    setup_commands: Optional[list[str]] = None,
) -> None:
    if not settings.api_key:
        redirect_to_login(settings)
        return

    run_until_complete(
        update_cloud_config_task(
            api_key=settings.api_key,
            base_api_url=settings.base_api_url,
            base_ws_url=settings.base_ws_url,
            cloud_config_uuid=cloud_config_uuid,
            github_org_name=github_org_name,
            github_repo_name=github_repo_name,
            setup_commands=setup_commands,
        )
    )


async def update_cloud_config_task(
    api_key: str,
    base_api_url: str,
    base_ws_url: str,
    cloud_config_uuid: str,
    github_org_name: str,
    github_repo_name: str,
    setup_commands: Optional[list[str]],
) -> None:
    graphql_client = GraphQLClient(api_key, base_api_url, base_ws_url)
    variables = {
        "cloudConfigUuid": cloud_config_uuid,
        "githubOrgName": github_org_name,
        "githubRepoName": github_repo_name,
        "setupCommands": setup_commands,
    }

    result = await graphql_client.execute(UPDATE_CLOUD_CONFIG_MUTATION, variables)
    click.echo(result)


@config_cli.command(hidden=True)
@click.option(
    "--email", required=True, help="Email of the user to upgrade to paid plan"
)
@use_settings
def upgrade_user_to_paid(
    settings: Settings,
    email: str,
) -> None:
    """Upgrade a user to trusted plan. Only available to @exponent.run admins."""
    if not settings.api_key:
        redirect_to_login(settings)
        return

    run_until_complete(
        upgrade_user_to_trusted_task(
            api_key=settings.api_key,
            base_api_url=settings.base_api_url,
            base_ws_url=settings.base_ws_url,
            email=email,
        )
    )


async def upgrade_user_to_trusted_task(
    api_key: str,
    base_api_url: str,
    email: str,
    base_ws_url: str,
) -> None:
    graphql_client = GraphQLClient(api_key, base_api_url, base_ws_url)
    variables = {
        "email": email,
    }
    result = await graphql_client.execute(UPGRADE_USER_TO_TRUSTED_MUTATION, variables)
    if "__typename" in result["upgradeUserToTrusted"]:
        if result["upgradeUserToTrusted"]["__typename"] == "Error":
            click.secho(f"Error: {result['upgradeUserToTrusted']['message']}", fg="red")
        else:
            click.secho(
                f"Successfully upgraded user {result['upgradeUserToTrusted']['email']} to {result['upgradeUserToTrusted']['plan']} plan",
                fg="green",
            )
    else:
        click.echo(result)


@config_cli.command()
@use_settings
def refresh_key(settings: Settings) -> None:
    """Refresh your API key."""
    if not settings.api_key:
        redirect_to_login(settings)
        return

    click.echo("Refreshing API key...")
    run_until_complete(
        refresh_api_key_task(
            api_key=settings.api_key,
            base_api_url=settings.base_api_url,
            base_ws_url=settings.base_ws_url,
        )
    )



from enum import Enum
from typing import Any, ClassVar, Literal, Optional, Union

from pydantic import BaseModel


class CommandType(str, Enum):
    THINKING = "thinking"
    FILE_READ = "file_read"
    SUMMARIZE = "summarize"
    STEP_OUTPUT = "step_output"
    PROTOTYPE = "prototype"
    DB_QUERY = "db_query"
    DB_GET_TABLE_NAMES = "db_get_table_names"
    DB_GET_TABLE_SCHEMA = "db_get_table_schema"


class CommandData(BaseModel):
    executable: ClassVar[bool]


class FileReadCommandData(CommandData):
    executable: ClassVar[bool] = True
    type: Literal[CommandType.FILE_READ] = CommandType.FILE_READ

    file_path: str
    language: str


class ThinkingCommandData(CommandData):
    executable: ClassVar[bool] = False
    type: Literal[CommandType.THINKING] = CommandType.THINKING

    content: str
    signature: Optional[str] = None


class PrototypeCommandData(CommandData):
    executable: ClassVar[bool] = True
    type: Literal[CommandType.PROTOTYPE] = CommandType.PROTOTYPE

    command_name: str
    # Structured data extracted from LLM output
    content_json: dict[str, Any]
    # Raw text extracted from LLM output
    content_raw: str
    # Rendered LLM output for frontend display
    content_rendered: str

    llm_command_name_override: Optional[str] = None

    @property
    def llm_command_name(self) -> str:
        return self.llm_command_name_override or self.command_name


# deprecated, use StepOutputCommandData instead
class SummarizeCommandData(CommandData):
    executable: ClassVar[bool] = True
    type: Literal[CommandType.SUMMARIZE] = CommandType.SUMMARIZE

    summary: str


class StepOutputCommandData(CommandData):
    executable: ClassVar[bool] = True
    type: Literal[CommandType.STEP_OUTPUT] = CommandType.STEP_OUTPUT

    step_output_raw: str


class DBQueryCommandData(CommandData):
    executable: ClassVar[bool] = True
    type: Literal[CommandType.DB_QUERY] = CommandType.DB_QUERY

    query: str


class DBGetTableNamesCommandData(CommandData):
    executable: ClassVar[bool] = True
    type: Literal[CommandType.DB_GET_TABLE_NAMES] = CommandType.DB_GET_TABLE_NAMES


class DBGetTableSchemaCommandData(CommandData):
    executable: ClassVar[bool] = True
    type: Literal[CommandType.DB_GET_TABLE_SCHEMA] = CommandType.DB_GET_TABLE_SCHEMA

    table_name: str


CommandDataType = Union[
    FileReadCommandData,
    ThinkingCommandData,
    PrototypeCommandData,
    SummarizeCommandData,  # deprecated
    DBQueryCommandData,
    DBGetTableNamesCommandData,
    DBGetTableSchemaCommandData,
    StepOutputCommandData,
]



HALT_CHAT_STREAM_MUTATION: str = """
  mutation HaltChatStream($chatUuid: UUID!) {
    haltChatStream(chatUuid: $chatUuid) {
      __typename
    }
  }
"""


SET_LOGIN_COMPLETE_MUTATION: str = """
  mutation SetLoginComplete {
    setLoginComplete {
      __typename
      ... on User {
        userApiKey
      }
      ... on UnauthenticatedError {
        message
      }
    }
  }
"""


UPGRADE_USER_TO_TRUSTED_MUTATION: str = """
  mutation UpgradeUserToTrusted($email: String!) {
    upgradeUserToTrusted(email: $email) {
      __typename
      ... on User {
        email
        plan
      }
      ... on Error {
        message
      }
    }
  }
"""


REFRESH_API_KEY_MUTATION = """
mutation RefreshApiKey {
    refreshApiKey {
        ... on User {
            userApiKey
        }
        ... on UnauthenticatedError {
            message
        }
    }
}
"""



import subprocess
from importlib.metadata import Distribution, PackageNotFoundError
from json import JSONDecodeError
from typing import Literal, Optional, Union, cast, Any
import sys
import click
from httpx import Client, HTTPError
from packaging.version import Version
import pkg_resources
import platform


def _is_pipx_install() -> bool:
    """Check if exponent was installed via pipx by examining sys.executable path."""
    return "pipx" in str(sys.executable)


def get_python_path() -> str:
    """Get the path to the Python interpreter."""
    try:
        return subprocess.check_output(["which", "python"]).decode().strip()
    except Exception:
        return "unknown"


def get_sys_executable() -> str:
    """Get the path to the Python interpreter."""
    return str(sys.executable)


def get_installed_version() -> Union[str, Literal["unknown"]]:
    """Get the installed version of exponent-run.

    Returns:
        The installed version of exponent-run if it can be determined, otherwise "unknown"
    """
    try:
        return Distribution.from_name("exponent-run").version
    except PackageNotFoundError as e:
        click.echo(f"Error reading version: {e}", err=True)
        return "unknown"


def get_installed_metadata() -> Union[Any, Literal["unknown"]]:
    """Get the installed metadata of exponent-run.

    Returns:
        The installed metadata of exponent-run if it can be determined, otherwise "unknown"
    """
    try:
        return Distribution.from_name("exponent-run").metadata
    except PackageNotFoundError as e:
        click.echo(f"Error reading metadata: {e}", err=True)
        return "unknown"


def get_installer() -> Union[str, Literal["unknown"]]:
    """Get the installer of exponent-run.

    Returns:
        The installer of exponent-run if it can be determined, otherwise "unknown"
    """
    try:
        return cast(
            str,
            pkg_resources.get_distribution("exponent-run").get_metadata("INSTALLER"),
        )
    except Exception:
        return "unknown"


def get_latest_pypi_exponent_version() -> Optional[str]:
    """Get the latest version of Exponent available on PyPI.

    Returns:
        The newest version of Exponent available on PyPI, or None if an error occurred.
    """
    try:
        return cast(
            str,
            (
                Client()
                .get("https://pypi.org/pypi/exponent-run/json")
                .json()["info"]["version"]
            ),
        )
    except (HTTPError, JSONDecodeError, KeyError):
        click.secho(
            "An unexpected error occurred communicating with PyPi, please check your network and try again.",
            fg="red",
        )
        return None


def check_exponent_version() -> Optional[tuple[str, str]]:
    """Check if there is a newer version of Exponent available on PyPI .

    Returns:
        None
    """

    installed_version = get_installed_version()
    if installed_version == "unknown":
        click.secho("Unable to determine current Exponent version.", fg="yellow")
        return None

    if (latest_version := get_latest_pypi_exponent_version()) and Version(
        latest_version
    ) > Version(installed_version):
        return installed_version, latest_version

    return None


def _get_pip_install():
    cmd = [
        sys.executable,
        "-m",
        "pip",
        "install",
        "--upgrade",
        "--upgrade-strategy",
        "only-if-needed",
        "exponent-run",
    ]
    return cmd


def upgrade_exponent(
    *,
    current_version: str,
    new_version: str,
    force: bool,
) -> None:
    """Upgrade Exponent to the passed in version.

    Args:
        current_version: The current version of Exponent.
        new_version: The new version of Exponent.
        force: Whether to force the upgrade without prompting for confirmation.

    Returns:
        None
    """
    upgrade_command = _get_pip_install()

    if platform.system() == "Windows":
        click.secho("Run this command to update:", fg="yellow")
        click.echo(f"{' '.join(upgrade_command)}")
        return

    if not force:
        click.secho(
            f"New version available: exponent-run=={new_version} (current: {current_version})\n"
            f"Update command: '{' '.join(upgrade_command)}'",
            fg="yellow",
            bold=True,
        )

        if not click.confirm("Update now?", default=True):
            click.secho("Aborted.", fg="red")
            return
    else:
        click.echo(f"Current version: {current_version}")
        click.echo(f"New version available: {new_version}")

    click.secho("Updating...", bold=True, fg="yellow")
    subprocess.check_call(upgrade_command)

    click.secho(f"Successfully upgraded Exponent to version {new_version}!", fg="green")



import asyncio
import os
import shutil
from dataclasses import dataclass
from typing import Optional, Callable

STDOUT_FD = 1
STDERR_FD = 2
MAX_TIMEOUT = 300


"""
Deprecated, use shell_streaming.py
"""


@dataclass
class ShellExecutionResult:
    output: str
    cancelled_for_timeout: bool
    exit_code: Optional[int]
    halted: bool = False


async def execute_shell(
    code: str,
    working_directory: str,
    timeout: int,
    should_halt: Optional[Callable[[], bool]] = None,
) -> ShellExecutionResult:
    timeout = min(timeout, MAX_TIMEOUT)

    shell_path = (
        os.environ.get("SHELL")
        or shutil.which("bash")
        or shutil.which("sh")
        or "/bin/sh"
    )

    process = await asyncio.create_subprocess_exec(
        shell_path,
        "-l",
        "-c",
        code,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
        cwd=working_directory,
    )

    exit_code = None
    output: list[tuple[int, str]] = []
    halted = False
    assert process.stdout
    assert process.stderr

    stdout_capture_task = asyncio.create_task(
        capture(process.stdout, STDOUT_FD, output)
    )
    stderr_capture_task = asyncio.create_task(
        capture(process.stderr, STDERR_FD, output)
    )

    async def capture_until_exit() -> int:
        nonlocal halted
        while True:
            if should_halt and should_halt():
                process.kill()
                halted = True
                break

            if process.returncode is not None:
                break

            await asyncio.sleep(0.1)

        return await process.wait()

    try:
        exit_code = await asyncio.wait_for(capture_until_exit(), timeout)
    except (TimeoutError, asyncio.TimeoutError):  # noqa: UP041
        process.kill()
    except Exception:
        raise
    finally:
        # Wait for capture tasks to complete naturally after streams are closed
        await asyncio.gather(
            stdout_capture_task, stderr_capture_task, return_exceptions=True
        )

    formatted_output = "".join([chunk for (_, chunk) in output]).strip() + "\n\n"

    return ShellExecutionResult(
        output=formatted_output,
        cancelled_for_timeout=exit_code is None and not halted,
        exit_code=exit_code,
        halted=halted,
    )


async def capture(
    stream: asyncio.StreamReader, fd: int, output: list[tuple[int, str]]
) -> None:
    while True:
        data = await stream.read(4096)
        if not data:
            break

        chunk = data.decode()
        output.append((fd, chunk))



from typing import Any, Optional

from pydantic import BaseModel
import traceback


class SerializableErrorInfo(BaseModel):
    message: str
    stack: list[str]
    cls_name: Optional[str]
    cause: Any
    context: Any

    def __str__(self) -> str:
        return self.to_string()

    def to_string(self) -> str:
        stack_str = "\nStack Trace:\n" + "".join(self.stack) if self.stack else ""
        cause_str = (
            "\nThe above exception was caused by the following exception:\n"
            + self.cause.to_string()
            if self.cause
            else ""
        )
        context_str = (
            "\nThe above exception occurred during handling of the following exception:\n"
            + self.context.to_string()
            if self.context
            else ""
        )

        return f"{self.message}{stack_str}{cause_str}{context_str}"


def serialize_error_info(error: BaseException) -> SerializableErrorInfo:
    return SerializableErrorInfo(
        message=str(error),
        stack=traceback.format_tb(error.__traceback__),
        cls_name=error.__class__.__name__,
        cause=serialize_error_info(error.__cause__) if error.__cause__ else None,
        context=serialize_error_info(error.__context__) if error.__context__ else None,
    )



import getpass
import os
import platform
from typing import Optional

from exponent.core.remote_execution.git import get_git_info
from exponent.core.remote_execution.languages import python_execution
from exponent.core.remote_execution.types import (
    SystemContextRequest,
    SystemContextResponse,
    SystemInfo,
)


EXPONENT_TXT_FILENAMES = [
    "exponent.txt",
]


def get_system_context(
    request: SystemContextRequest, working_directory: str
) -> SystemContextResponse:
    return SystemContextResponse(
        correlation_id=request.correlation_id,
        exponent_txt=_read_exponent_txt(working_directory),
        system_info=get_system_info(working_directory),
    )


def get_system_info(working_directory: str) -> SystemInfo:
    return SystemInfo(
        name=getpass.getuser(),
        cwd=working_directory,
        os=platform.system(),
        shell=_get_user_shell(),
        git=get_git_info(working_directory),
        python_env=python_execution.get_python_env_info(),
    )


def _read_exponent_txt(working_directory: str) -> Optional[str]:
    for filename in EXPONENT_TXT_FILENAMES:
        file_path = os.path.join(working_directory, filename.lower())

        if os.path.exists(file_path):
            with open(file_path) as f:
                return f.read()

    return None


def _get_user_shell() -> str:
    return os.environ.get("SHELL", "bash")



import asyncio
import os
import shutil
from exponent.core.remote_execution.languages.types import (
    StreamedOutputPiece,
    ShellExecutionResult,
)
from typing import Optional, Callable, AsyncGenerator, Union


STDOUT_FD = 1
STDERR_FD = 2
MAX_TIMEOUT = 300


async def read_stream(
    stream: asyncio.StreamReader, fd: int, output: list[tuple[int, str]]
) -> AsyncGenerator[StreamedOutputPiece, None]:
    while True:
        try:
            data = await stream.read(4096)
            if not data:
                break
            chunk = data.decode()
            output.append((fd, chunk))
            yield StreamedOutputPiece(content=chunk)
        except asyncio.CancelledError:
            raise
        except Exception:
            break


async def execute_shell_streaming(
    code: str,
    working_directory: str,
    timeout: int,
    should_halt: Optional[Callable[[], bool]] = None,
) -> AsyncGenerator[Union[StreamedOutputPiece, ShellExecutionResult], None]:
    timeout_seconds = min(timeout, MAX_TIMEOUT)

    shell_path = (
        os.environ.get("SHELL")
        or shutil.which("bash")
        or shutil.which("sh")
        or "/bin/sh"
    )

    process = await asyncio.create_subprocess_exec(
        shell_path,
        "-l",
        "-c",
        code,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
        cwd=working_directory,
    )

    exit_code = None
    output: list[tuple[int, str]] = []
    halted = False
    timed_out = False
    assert process.stdout
    assert process.stderr

    async def monitor_halt() -> None:
        nonlocal halted
        while True:
            if should_halt and should_halt():
                process.kill()
                halted = True
                break
            if process.returncode is not None:
                break
            await asyncio.sleep(0.1)

    def on_timeout() -> None:
        nonlocal timed_out
        timed_out = True
        process.kill()

    try:
        halt_task = asyncio.create_task(monitor_halt()) if should_halt else None
        timeout_handle = asyncio.get_running_loop().call_later(
            timeout_seconds, on_timeout
        )

        # Stream stdout and stderr concurrently using wait
        stdout_gen = read_stream(process.stdout, STDOUT_FD, output)
        stderr_gen = read_stream(process.stderr, STDERR_FD, output)

        stdout_task = asyncio.create_task(stdout_gen.__anext__())
        stderr_task = asyncio.create_task(stderr_gen.__anext__())
        pending = {stdout_task, stderr_task}

        while pending:
            done, pending = await asyncio.wait(
                pending, return_when=asyncio.FIRST_COMPLETED
            )

            for task in done:
                try:
                    piece = await task
                    yield piece

                    # Schedule next read from the same stream
                    if task is stdout_task and not process.stdout.at_eof():
                        stdout_task = asyncio.create_task(stdout_gen.__anext__())
                        pending.add(stdout_task)
                    elif task is stderr_task and not process.stderr.at_eof():
                        stderr_task = asyncio.create_task(stderr_gen.__anext__())
                        pending.add(stderr_task)
                except StopAsyncIteration:
                    continue

        exit_code = await process.wait()
        timeout_handle.cancel()

    except asyncio.CancelledError:
        process.kill()
        raise
    finally:
        if halt_task and not halt_task.done():
            halt_task.cancel()
            try:
                await halt_task
            except asyncio.CancelledError:
                pass

    formatted_output = "".join([chunk for (_, chunk) in output]).strip() + "\n\n"

    yield ShellExecutionResult(
        output=formatted_output,
        cancelled_for_timeout=timed_out,
        exit_code=None if timed_out else exit_code,
        halted=halted,
    )



import os
from collections.abc import Callable
from pathlib import Path
from typing import Optional, Union, cast
from anyio import Path as AsyncPath

import pygit2
from pygit2 import Tree
from pygit2.repository import Repository
from pygit2.enums import DiffOption
from gitignore_parser import (
    IgnoreRule,
    handle_negation,
    parse_gitignore,
    rule_from_pattern,
)

from exponent.core.remote_execution.types import (
    GitInfo,
)

GIT_OBJ_COMMIT = 1


async def git_file_walk(
    repo: Repository,
    directory: str,
) -> list[str]:
    """
    Walk through a directory and return all file paths, respecting .gitignore and additional ignore patterns.
    """
    tree = get_git_subtree_for_dir(repo, directory)

    if not tree:
        return []

    # diff to the empty tree to see all files
    tracked_diff = tree.diff_to_tree()

    tracked_files = [delta.new_file.path for delta in tracked_diff.deltas]

    # Find untracked files relative to the root
    untracked_diff = repo.diff(flags=DiffOption.INCLUDE_UNTRACKED)
    untracked_files_from_root = [
        AsyncPath(delta.new_file.path) for delta in untracked_diff.deltas
    ]

    # Current working directory relative to the repo root
    dir_path = await AsyncPath(directory).resolve()
    repo_path = await AsyncPath(repo.workdir).resolve()

    if repo_path == dir_path:
        relative_directory = str(repo_path)
    else:
        relative_directory = str(dir_path.relative_to(repo_path))

    # Resolve all untracked files that are within the current working directory
    untracked_files = []
    for untracked_file in untracked_files_from_root:
        if not untracked_file.is_relative_to(relative_directory):
            continue

        untracked_files.append(str(untracked_file.relative_to(relative_directory)))

    # Combine both as sets to remove duplicates
    return list(set(tracked_files) | set(untracked_files))


def get_repo(working_directory: str) -> Optional[Repository]:
    try:
        return Repository(working_directory)
    except pygit2.GitError:
        return None


def get_git_info(working_directory: str) -> Optional[GitInfo]:
    try:
        repo = Repository(working_directory)
    except pygit2.GitError:
        return None

    return GitInfo(
        branch=_get_git_branch(repo) or "<unknown branch>",
        remote=_get_git_remote(repo),
    )


def get_tracked_files_in_dir(
    repo: Repository,
    dir: Union[str, Path],
    filter_func: Optional[Callable[[str], bool]] = None,
) -> list[str]:
    rel_path = get_path_relative_to_repo_root(repo, dir)
    dir_tree = get_git_subtree_for_dir(repo, dir)
    entries: list[str] = []
    if not dir_tree:
        return entries
    for entry in dir_tree:
        if not entry.name:
            continue
        entry_path = str(Path(f"{repo.workdir}/{rel_path}/{entry.name}"))
        if entry.type_str == "tree":
            entries.extend(get_tracked_files_in_dir(repo, entry_path, filter_func))
        elif entry.type_str == "blob":
            if not filter_func or filter_func(entry.name):
                entries.append(entry_path)
    return entries


def get_git_subtree_for_dir(repo: Repository, dir: Union[str, Path]) -> Optional[Tree]:
    rel_path = get_path_relative_to_repo_root(repo, dir)

    try:
        head_commit = repo.head.peel(GIT_OBJ_COMMIT)
    except pygit2.GitError:
        # If the repo is empty, then the head commit will not exist
        return None
    head_tree: Tree = head_commit.tree

    if rel_path == Path("."):
        # If the relative path is the root of the repo, then
        # the head_tree is what we want. Note we do this because
        # Passing "." or "" as the path into the tree will raise.
        return head_tree
    return cast(Tree, head_tree[str(rel_path)])


def get_path_relative_to_repo_root(repo: Repository, path: Union[str, Path]) -> Path:
    path = Path(path).resolve()
    return path.relative_to(Path(repo.workdir).resolve())


def get_local_commit_hash() -> str:
    try:
        # Open the repository (assumes the current working directory is within the git repo)
        repo = Repository(os.getcwd())

        # Get the current HEAD commit
        head = repo.head

        # Get the commit object and return its hash as a string
        return str(repo[head.target].id)
    except pygit2.GitError:
        return "unknown-local-commit"


def _get_git_remote(repo: Repository) -> Optional[str]:
    if repo.remotes:
        return str(repo.remotes[0].url)
    return None


def _get_git_branch(repo: Repository) -> Optional[str]:
    try:
        # Look for HEAD file in the .git directory
        head_path = os.path.join(repo.path, "HEAD")
        with open(head_path) as head_file:
            # The HEAD file content usually looks like: 'ref: refs/heads/branch_name'
            head_content = head_file.read().strip()
            if head_content.startswith("ref:"):
                return head_content.split("refs/heads/")[-1]
            else:
                return None
    except Exception:  # noqa: BLE001
        return None


class GitIgnoreHandler:
    def __init__(
        self, working_directory: str, default_ignores: Optional[list[str]] = None
    ):
        self.checkers = {}

        if default_ignores:
            self.checkers[working_directory] = self._parse_ignore_extra(
                working_directory, default_ignores
            )

    async def read_ignorefile(self, path: str) -> None:
        new_ignore = await self._get_ignored_checker(path)

        if new_ignore:
            self.checkers[path] = new_ignore

    def filter(
        self,
        relpaths: list[str],
        root: str,
    ) -> list[str]:
        result = []

        for relpath in relpaths:
            if relpath.startswith(".git"):
                continue

            path = os.path.join(root, relpath)

            if self.is_ignored(path):
                continue

            result.append(relpath)

        return result

    def is_ignored(self, path: str) -> bool:
        return any(
            self.checkers[dp](path)
            for dp in self.checkers
            if self._is_subpath(path, dp)
        )

    def _parse_ignore_extra(
        self, working_directory: str, ignore_extra: list[str]
    ) -> Callable[[str], bool]:
        rules: list[IgnoreRule] = []

        for pattern in ignore_extra:
            if (
                rule := rule_from_pattern(pattern, base_path=working_directory)
            ) is not None:
                rules.append(rule)

        def rule_handler(file_path: str) -> bool:
            nonlocal rules
            return bool(handle_negation(file_path, rules))

        return rule_handler

    async def _get_ignored_checker(
        self, dir_path: str
    ) -> Optional[Callable[[str], bool]]:
        new_ignore = self._parse_gitignore(dir_path)

        existing_ignore = self.checkers.get(dir_path)

        if existing_ignore and new_ignore:
            return self._or(new_ignore, existing_ignore)

        return new_ignore or existing_ignore

    @staticmethod
    def _parse_gitignore(directory: str) -> Optional[Callable[[str], bool]]:
        gitignore_path = os.path.join(directory, ".gitignore")

        if os.path.isfile(gitignore_path):
            return cast(Callable[[str], bool], parse_gitignore(gitignore_path))

        return None

    @staticmethod
    def _or(
        a: Callable[[str], bool], b: Callable[[str], bool]
    ) -> Callable[[str], bool]:
        def or_handler(file_path: str) -> bool:
            return a(file_path) or b(file_path)

        return or_handler

    @staticmethod
    def _is_subpath(path: str, parent: str) -> bool:
        """
        Check if a path is a subpath of another path.
        """
        return os.path.commonpath([path, parent]) == parent



from dataclasses import dataclass
from typing import Optional


@dataclass
class StreamedOutputPiece:
    content: str


@dataclass
class ShellExecutionResult:
    output: str
    cancelled_for_timeout: bool
    exit_code: Optional[int]
    halted: bool = False


@dataclass
class PythonExecutionResult:
    output: str
    halted: bool = False
